/* load.c (emx+gcc) */

/* To speed up compilation of many small files with GCC, run the
   following command:

       load gcc cpp cc1 as

   This will load the GCC omponents into memory. */

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <errno.h>
#include <process.h>
#include <sys/ptrace.h>

static int load (const char *prog)
{
  int pid;

  pid = spawnl (P_DEBUG | P_WINDOWED | P_NOSESSION, prog, NULL);
  if (pid < 0)
    perror (prog);
  return (pid);
}


static void stop (int pid)
{
  if (pid >= 0)
    ptrace (PTRACE_EXIT, pid, 0, 0);
}


static void usage (void)
{
  fputs ("Usage: load [-f] [-t<minutes>] <program>...\n", stderr);
  exit (1);
}


int main (int argc, char **argv)
{
  int c, i, n, *pids;
  int forever, minutes;
  char buf[2], *p;

  if (_osmode == DOS_MODE)
    {
      fputs ("There's no point in using this program under DOS\n", stderr);
      return (1);
    }
  forever = 0; minutes = 0;
  while ((c = getopt (argc, argv, "ft:")) != EOF)
    switch (c)
      {
      case 'f':
        forever = 1;
        break;
      case 't':
        errno = 0;
        minutes = (int)strtol (optarg, &p, 10);
        if (minutes < 1 || errno != 0 || *p != 0)
          usage ();
        break;
      default:
        usage ();
      }
  argc -= optind;
  argv += optind;
  if (argc < 1)
    usage ();
  pids = alloca (argc * sizeof (*pids));
  n = 0;
  for (i = 0; i < argc; ++i)
    {
      pids[i] = load (argv[i]);
      if (pids[i] >= 0)
        ++n;
    }
  if (n <= 0)
    return (1);
  if (forever)
    for (;;)
      sleep (4000000);
  else if (minutes != 0)
    sleep (minutes * 60);
  else
    {
      printf ("Type RETURN to quit: "); fflush (stdout);
      fgets (buf, sizeof (buf), stdin);
    }
  for (i = 1; i < argc; ++i)
    stop (pids[i]);
  return (0);
}
